import json
import math

from ph4_walkingpad.utils import defval


def calories_bmr(weight: float, height: float, age: float, male=True) -> float:
    """http://www.shapesense.com/fitness-exercise/calculators/net-versus-gross-calorie-burn-conversion-calculator.shtml"""
    if male:
        return 13.75 * weight + 5 * (height*100) - 6.76 * age + 66
    else:
        return 9.56 * weight + 1.85 * (height*100) - 4.68 * age + 655


def calories_rmrcb_minute(weight: float, height: float, age: float, male=True) -> float:
    """http://www.shapesense.com/fitness-exercise/calculators/net-versus-gross-calorie-burn-conversion-calculator.shtml
    per minute"""
    return 1/60 * (calories_bmr(weight, height, age, male) * 1.1 / 24)


def calories_walk_minute(speed: float, weight: float, height: float):
    """Cal per minute = (0.035  X body weight in kg) + ((Velocity in m/s ^ 2) / Height in m))
        X (0.029) X (body weight in kg)"""
    vel = speed * 1000 / 3600  # to kmph
    return (0.035 * weight) + ((vel * vel) / height) * (0.029 * weight)


def calories_walk2_minute(speed: float, weight: float, deg: float):
    """
    http://www.shapesense.com/fitness-exercise/calculators/walking-calorie-burn-calculator.shtml
    Valid only for speeds in range 1 - 7.5 kmph
    deg: elevation, 5% elev = 0.05
    """
    speed_matrix = [
        [0.0251, - 0.2157, + 0.7888, + 1.2957],
        [0.0244, - 0.2079, + 0.8053, + 1.3281],
        [0.0237, - 0.2000, + 0.8217, + 1.3605],
        [0.0230, - 0.1922, + 0.8382, + 1.3929],
        [0.0222, - 0.1844, + 0.8546, + 1.4253],
        [0.0215, - 0.1765, + 0.8710, + 1.4577],
        [0.0171, - 0.1062, + 0.6080, + 1.8600],
        [0.0184, - 0.1134, + 0.6566, + 1.9200],
        [0.0196, - 0.1205, + 0.7053, + 1.9800],
        [0.0208, - 0.1277, + 0.7539, + 2.0400],
        [0.0221, - 0.1349, + 0.8025, + 2.1000],
    ]
    if deg <= 0.06:
        matrix_idx = min(len(speed_matrix) - 1, int(round(int(max(-5, deg * 100.0)) + 5.0)))
        return 1/60. * sum([speed_matrix[matrix_idx][i] * (speed ** (3-i)) for i in range(4)]) * weight

    mpm = speed * 1000 / 60
    return 1/60. * ((0.1 * mpm + 1.8 * mpm * deg + 3.5) * weight * 60 * 5 / 1000)


class Profile:
    def __init__(self, pid=None, male=True, age=0, weight=0, height=0, token=None, did=None, email=None, password=None,
                 password_md5=None):
        self.pid = pid
        self.male = male
        self.age = age
        self.weight = weight
        self.height = height
        self.token = token
        self.did = did
        self.email = email
        self.password = password
        self.password_md5 = password_md5

    def load_from(self, js):
        self.pid = defval(js, 'id')
        self.male = defval(js, 'male')
        self.age = defval(js, 'age')
        self.weight = defval(js, 'weight')
        self.height = defval(js, 'height')
        self.token = defval(js, 'token')
        self.did = defval(js, 'did')
        self.email = defval(js, 'email')
        self.password = defval(js, 'password')
        self.password_md5 = defval(js, 'password_md5')

    def dump(self):
        js = {
            'id': self.pid,
            'male': self.male,
            'age': self.age,
            'weight': self.weight,
            'height': self.height,
            'token': self.token,
            'did': self.did,
            'email': self.email,
            'password': self.password,
            'password_md5': self.password_md5,
        }
        return js

    @staticmethod
    def from_data(data):
        m = Profile()
        m.load_from(data)
        return m

    def __str__(self):
        return 'Profile(pid=%s, male=%s, age=%s, weight=%s, height=%s)' \
               % (self.pid, self.male, self.age, self.weight, self.height)




import requests
import logging
import json
import hashlib

logger = logging.getLogger(__name__)


def upload_record(tok, did, cal, timex, dur, distance, step, **kwargs):
    """
    Uploads record to your account
    tok = your JWT token obtained from the app.
    did = device ID, MAC address ff:ff:ff:ff:ff:ff
    """
    url = 'https://eu.app.walkingpad.com/user/api/v2/record'
    cookies = {'user': tok}
    js = {'did': did, 'cal': cal, 'time': timex, 'dur': dur, 'distance': distance, 'step': step,
          'sid': None, 'model': 'A1'}
    logger.info('Upload record: %s' % (json.dumps(js, indent=2)),)
    return requests.post(url, json=js, cookies=cookies, **kwargs)


def get_records(tok, page=1, per_page=10000, timestamp=None, **kwargs):
    url = 'https://eu.app.walkingpad.com/user/api/v2/record?page=%d&per_page=%d' % (page, per_page)
    if timestamp:
        url += '&timestamp=%d' % timestamp
    cookies = {'user': tok}
    return requests.get(url, cookies=cookies, **kwargs)


def login(email, password=None, password_md5=None, **kwargs):
    """
    Logs in to the walking pad service, returns tuple (jwt-token, response)
    """
    url = 'https://eu.app.walkingpad.com/user/api/v2/login'
    password_md5 = password_md5 if password_md5 else hashlib.md5(password.encode('utf8')).hexdigest()
    js = {'email': email, 'password': password_md5}

    r = requests.post(url, json=js, **kwargs)
    r.raise_for_status()
    return r.cookies.get_dict()['user'], r





import json
import logging
from typing import Optional

from ph4_walkingpad.profile import Profile, calories_walk2_minute, calories_rmrcb_minute
from ph4_walkingpad.reader import reverse_file

logger = logging.getLogger(__name__)


class StatsAnalysis:
    def __init__(self, profile=None, profile_file=None, stats_file=None):
        self.profile_file = profile_file
        self.stats_file = stats_file
        self.profile = profile

        self.last_record = None
        self.loaded_margins = []

    def load_profile(self):
        self.profile = Profile(age=30, male=True, weight=80, height=180)  # some random average person
        if self.profile_file:
            with open(self.profile_file, 'r') as fh:
                dt = json.load(fh)
                self.profile = Profile.from_data(dt)

    def load_stats(self, limit=None, collect_details=False):
        for margins in self.parse_stats(limit, collect_details=collect_details):
            self.loaded_margins.append(margins)

    def feed_records(self):
        """Feed records from stats file in reversed order, one record per entry"""
        if not self.stats_file:
            return

        with open(self.stats_file) as fh:
            reader = reverse_file(fh)
            for line in reader:
                if line is None:
                    return
                if not line:
                    continue

                try:
                    js = json.loads(line)
                except Exception as e:
                    continue

                yield js

    def analyze_records_margins(self, records, limit=None, collect_details=False):
        # Load margins - boundary speed changes. In order to determine segments of the same speed.
        last_rec = None
        last_rec_diff = None
        in_record = False
        num_done = 0
        margins = []
        sub_records = []

        for js in records:
            if not self.last_record:
                self.last_record = js

            if js['speed'] != 0:
                in_record = True

            if not last_rec_diff or not in_record:
                last_rec_diff = js
                last_rec = js
                sub_records = []
                if not in_record:
                    margins = [js]
                else:
                    margins.append(js)
                continue

            time_diff = last_rec['time'] - js['time']
            steps_diff = last_rec['steps'] - js['steps']
            dist_diff = last_rec['dist'] - js['dist']
            rtime_diff = last_rec['rec_time'] - js['rec_time']
            time_to_rtime = abs(time_diff - rtime_diff)
            js['_ldiff'] = [time_diff, steps_diff, dist_diff, rtime_diff, time_to_rtime]

            breaking = time_diff < 0 or steps_diff < 0 or dist_diff < 0 or rtime_diff < 0 or time_to_rtime > 5*60
            stats_changed = False

            if in_record and collect_details:
                sub_records.append(dict(js))

            if breaking:
                if margins:
                    mm = margins[-1]
                    mm['_breaking'] = breaking

            if (in_record or not breaking) \
                    and (last_rec_diff['speed'] != js['speed']
                    or (breaking and last_rec_diff['speed'] != 0)
                    or (js['speed'] == 0 and js['time'] == 0)):

                js['_breaking'] = breaking
                js_src = js if not breaking else last_rec
                if margins:
                    mm = margins[-1]
                    mm['_segment_time'] = last_rec_diff['time'] - js_src['time']
                    mm['_segment_rtime'] = last_rec_diff['rec_time'] - js_src['rec_time']
                    mm['_segment_dist'] = last_rec_diff['dist'] - js_src['dist']
                    mm['_segment_steps'] = last_rec_diff['steps'] - js_src['steps']
                    if collect_details:
                        mm['_records'] = sub_records[:-1]
                        sub_records = [dict(js)]

                margins.append(js)
                stats_changed = True
                last_rec_diff = js

            if (stats_changed and js['speed'] == 0 and js['time'] == 0) or breaking:
                # print("done", breaking, time_to_rtime, time_diff, steps_diff, dist_diff, rtime_diff, js)
                # logger.info(json.dumps(margins, indent=2))
                if margins:
                    yield margins
                    num_done += 1
                    if limit and num_done >= limit:
                        return

                margins = [js]
                in_record = False
                last_rec_diff = js

            # last inst.
            last_rec = js

    def parse_stats(self, limit=None, collect_details=False):
        gen = self.feed_records()
        return self.analyze_records_margins(gen, limit, collect_details=collect_details)

    def comp_calories(self, margins):
        # logger.debug(json.dumps(margins, indent=2))
        # Calories segment computation
        if not self.profile:
            logger.debug('No profile loaded')
            return

        calorie_acc = []
        calorie_acc_net = []
        for exp in margins:
            if '_segment_time' not in exp:
                continue

            el_time = exp['_segment_time']
            speed = exp['speed'] / 10.

            ccal = (el_time / 60) * calories_walk2_minute(speed, self.profile.weight, 0.00)
            ccal_net = ccal - (el_time / 60) * calories_rmrcb_minute(self.profile.weight, self.profile.height,
                                                                     self.profile.age, self.profile.male)

            logger.info('Calories for time %5s, speed %4.1f, seg time: %4s, dist: %5.2f, steps: %5d, '
                        'cal: %7.2f, ncal: %7.2f'
                        % (exp['time'], speed, el_time, exp['_segment_dist'] / 100., exp['_segment_steps'],
                           ccal, ccal_net))

            calorie_acc.append(ccal)
            calorie_acc_net.append(ccal_net)

        logger.info('Calories burned so far this walk: %7.2f kcal, %7.2f kcal net'
                    % (sum(calorie_acc), sum(calorie_acc_net)))
        return calorie_acc, calorie_acc_net

    def load_last_stats(self, count=1):
        self.load_stats(count)
        if self.loaded_margins:
            logger.debug('Loaded margins: %s' % (json.dumps(self.loaded_margins[0], indent=2),))
            return self.comp_calories(self.loaded_margins[0])

    def remove_records(self, margins):
        ret = []
        for recs in margins:
            nrecs = [dict(x) for x in recs]
            for rec in nrecs:
                rec['_records'] = None
            ret.append(nrecs)
        return ret



import logging
import sys
import re


def try_fnc(fnc):
    try:
        return fnc()
    except:
        pass


def defval(js, key, default=None):
    return js[key] if key in js else default


def setup_logging():
    log = logging.getLogger(__name__)
    log.setLevel(logging.DEBUG)
    h = logging.StreamHandler(sys.stdout)
    h.setLevel(logging.DEBUG)
    log.addHandler(h)
    return log


def parse_time_string(timx):
    mt = re.match(r'^(?:([\d.]+)(?:(?<=\d)h)?)?\s*\b(?:([\d.]+)(?:(?<=\d)m)?)?\s*\b(?:([\d.]+)(?:(?<=\d)s)?)?\s*$', timx.strip())
    if ':' in timx:
        parts = timx.split(':')
        if len(parts) < 2 or len(parts) > 3:
            raise ValueError('Time format error')

        if len(parts) == 2:
            return float(parts[0]) * 3600 + float(parts[1]) * 60
        else:
            return float(parts[0]) * 3600 + float(parts[1]) * 60 + float(parts[2])

    elif mt:
        return float(mt.group(1) or 0) * 3600 + float(mt.group(2) or 0) * 60 + float(mt.group(3) or 0)

    else:
        return float(timx)



import sys

from setuptools import setup
from setuptools import find_packages

version = '0.0.5'

# Please update tox.ini when modifying dependency version requirements
install_requires = [
    'setuptools>=1.0',
    'six',
    'future',
    'asyncio',
    'coloredlogs',
    'bleak>=0.14.1',
    'ph4-acmd2==0.0.5',
    'blessed',
    'requests',
    'aioconsole==0.3.1',
]

dev_extras = [
    'nose',
    'pep8',
    'tox',
    'pypandoc',
]

docs_extras = [
    'Sphinx>=1.0',  # autodoc_member_order = 'bysource', autodoc_default_flags
    'sphinx_rtd_theme',
    'sphinxcontrib-programoutput',
]

try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
    long_description = long_description.replace("\r", '')

except(IOError, ImportError):
    import io
    with io.open('README.md', encoding="utf-8") as f:
        long_description = f.read()

setup(
    name='ph4-walkingpad',
    version=version,
    description='Walkingpad A1 controller via Bluetooth LE',
    long_description=long_description,
    url='https://github.com/ph4r05/ph4-walkingpad',
    author='Dusan Klinec',
    author_email='dusan.klinec@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'dev': dev_extras,
        'docs': docs_extras,
    },
    entry_points={
        'console_scripts': [
            'ph4-walkingpad-ctl = ph4_walkingpad.main:main',
            'ph4-cal = ph4_walkingpad.cal:main',
        ],
    }
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import json
import logging
from typing import Union

import coloredlogs

from ph4_walkingpad.profile import Profile, calories_walk2_minute, calories_rmrcb_minute
from ph4_walkingpad.utils import parse_time_string

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.INFO)


class CalMeter:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = None
        self.profile = None

    def work(self):
        speed = self.args.speed
        timx = self.args.time  # type: Union[str,int,float]
        dist = self.args.dist

        if timx:
            old_time = timx
            if isinstance(timx, str):
                timx = parse_time_string(timx)
                logger.debug('Time %s recomputed to %s seconds' % (old_time, timx,))

        if not speed and timx and dist:
            speed = dist / (timx/3600)

        if not timx and speed and dist:
            timx = 3600 * (dist / speed)

        if not dist and timx and speed:
            dist = (timx / 3600) * speed

        if not speed:
            raise ValueError('Missing speed')

        hrs = int(timx / 3600)
        mnts = int((timx - hrs * 3600) / 60)
        scnds = int(timx - hrs*3600 - mnts*60)

        ccal = (timx / 60) * calories_walk2_minute(speed, self.profile.weight, 0.00)
        ccal_net = ccal - (timx / 60) * calories_rmrcb_minute(self.profile.weight, self.profile.height,
                                                              self.profile.age, self.profile.male)

        print("Speed: %4.1f km/h, dist: %5.2f km, time: %5s s = %02s :%02s :%02s, cal: %7.2f, ncal: %7.2f"
              % (speed, dist, timx, hrs, mnts, scnds, ccal, ccal_net))

        ccal = (timx / 60) * calories_walk2_minute(speed, self.profile.weight, 0.00)
        ccal_net = ccal - (timx / 60) * calories_rmrcb_minute(self.profile.weight, self.profile.height,
                                                              self.profile.age, self.profile.male)

    # noinspection DuplicatedCode
    def load_profile(self):
        self.profile = Profile(age=30, male=True, weight=80, height=180)  # some random average person
        if self.args.profile:
            with open(self.args.profile, 'r') as fh:
                dt = json.load(fh)
                self.profile = Profile.from_data(dt)

        if self.args.weight:
            self.profile.weight = self.args.weight
        if self.args.height:
            self.profile.height = self.args.height
        if self.args.female:
            self.profile.male = False
        if self.args.age:
            self.profile.age = self.args.age

    def main(self):
        parser = self.argparser()
        self.args = parser.parse_args()

        if self.args.debug:
            coloredlogs.install(level=logging.DEBUG)
        else:
            coloredlogs.install(level=logging.INFO)

        self.load_profile()
        self.work()

    def argparser(self):
        parser = argparse.ArgumentParser(description='ph4 Simple walking calories computation')

        parser.add_argument('--debug', dest='debug', action='store_const', const=True,
                            help='enables debug mode')
        parser.add_argument('-p', '--profile', dest='profile',
                            help='Profile JSON file')
        parser.add_argument('-w', '--weight', dest='weight', type=float, default=None,
                            help='Weight of the person in kg')
        parser.add_argument('-H', '--height', dest='height', type=float, default=None,
                            help='Height of the person in m')
        parser.add_argument('-a', '--age', dest='age', type=float, default=None,
                            help='Age of the person')
        parser.add_argument('-f', '--female', dest='female', action='store_const', const=True,
                            help='Person is a female')
        parser.add_argument('-s', '--speed', dest='speed', type=float,
                            help='Speed of the walk in km/h')
        parser.add_argument('-t', '--time', dest='time',
                            help='Time of the walk in hours h:m:s or 1h 2m 3s')
        parser.add_argument('-d', '--dist', dest='dist',
                            help='Distance of the walk in km')
        return parser


def main():
    br = CalMeter()
    br.main()


if __name__ == '__main__':
    main()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
import binascii
import logging
import time
import platform

import bleak
from bleak import discover
from bleak import BleakClient

# typing
if False:
    from bleak.backends.device import BLEDevice


logger = logging.getLogger(__name__)


class Scanner:
    UUIDS = [
        "0000180a-0000-1000-8000-00805f9b34fb",
        "00010203-0405-0607-0809-0a0b0c0d1912",
        "0000fe00-0000-1000-8000-00805f9b34fb",
    ]

    BLEAK_KWARGS = {
        "service_uuids": UUIDS
    }

    def __init__(self):
        self.devices_dict = {}
        self.devices_list = []
        self.receive_data = []
        self.walking_belt_candidates = []  # type: list[BLEDevice]

    @staticmethod
    def is_darwin():
        try:
            return platform.system().lower() == 'darwin'
        except:
            return False

    @staticmethod
    def get_bleak_kwargs():
        return Scanner.BLEAK_KWARGS if Scanner.is_darwin() else {}

    async def scan(self, timeout=3.0):
        kwargs = Scanner.get_bleak_kwargs()
        logger.info("Scanning for peripherals...")
        logger.debug("Scanning kwargs: %s" % (kwargs,))
        scanner = bleak.BleakScanner(**kwargs)
        dev = await scanner.discover(timeout=timeout, **kwargs)
        for i in range(len(dev)):
            # Print the devices discovered
            info_str = ', '.join(["[%2d]" % i, str(dev[i].address), str(dev[i].name), str(dev[i].metadata["uuids"])])
            logger.info("Device: %s" % info_str)

            # Put devices information into list
            self.devices_dict[dev[i].address] = []
            self.devices_dict[dev[i].address].append(dev[i].name)
            self.devices_dict[dev[i].address].append(dev[i].metadata["uuids"])
            self.devices_list.append(dev[i].address)

            if 'walkingpad' in dev[i].name.lower():
                self.walking_belt_candidates.append(dev[i])


class WalkingPad:
    MODE_STANDBY = 2
    MODE_MANUAL = 1
    MODE_AUTOMAT = 0

    PREFS_MAX_SPEED = 3
    PREFS_START_SPEED = 4
    PREFS_START_INTEL = 5
    PREFS_SENSITIVITY = 6
    PREFS_DISPLAY = 7
    PREFS_CHILD_LOCK = 9
    PREFS_UNITS = 8
    PREFS_TARGET = 1

    TARGET_NONE = 0
    TARGET_DIST = 1
    TARGET_CAL = 2
    TARGET_TIME = 3

    BUTTON_None = 0
    BUTTON_Down = 4
    BUTTON_Stop = 3
    BUTTON_Up = 2
    BUTTON_long_mode = -6
    BUTTON_up = -4
    BUTTON_mode = -6

    PAYLOADS_255 = [
        [247, 165, 96, 74, 77, 147, 113, 41, 201, 253],
        [247, 165, 96, 74, 58, 60, 113, 41, 95, 253],
        [247, 165, 96, 74, 15, 165, 113, 41, 157, 253],
        [247, 165, 96, 74, 21, 129, 113, 41, 127, 253],
        [247, 165, 96, 74, 45, 189, 115, 171, 87, 253],
        [247, 165, 96, 74, 49, 42, 113, 41, 68, 253],
        [247, 165, 96, 74, 58, 60, 113, 41, 95, 253],
        [247, 165, 96, 74, 77, 147, 113, 41, 201, 253],
    ]

    @staticmethod
    def int2byte(val, width=3):
        return [(val >> (8 * (width - 1 - i)) & 0xff) for i in range(width)]

    @staticmethod
    def byte2int(val, width=3):
        return sum([(val[i] << (8 * (width - 1 - i))) for i in range(width)])

    @staticmethod
    def fix_crc(cmd):
        cmd[-2] = sum(cmd[1:-2]) % 256
        return cmd


class WalkingPadCurStatus:
    def __init__(self):
        self.raw = None
        self.dist = 0
        self.time = 0
        self.steps = 0
        self.speed = 0
        self.controller_button = 0
        self.app_speed = 0
        self.belt_state = 0
        self.manual_mode = 0
        self.rtime = 0

    def load_from(self, cmd):
        self.raw = bytearray(cmd)
        self.belt_state = cmd[2]
        self.speed = cmd[3]
        self.manual_mode = cmd[4]
        self.time = WalkingPad.byte2int(cmd[5:])
        self.dist = WalkingPad.byte2int(cmd[8:])
        self.steps = WalkingPad.byte2int(cmd[11:])
        self.app_speed = cmd[14]  # / 30
        self.controller_button = cmd[16]
        self.rtime = time.time()

    @staticmethod
    def check_type(cmd):
        return bytes(cmd[0:2]) == bytes([248, 162])

    @staticmethod
    def from_data(cmd):
        if not WalkingPadCurStatus.check_type(cmd):
            raise ValueError('Incorrect message type, could not parse')
        m = WalkingPadCurStatus()
        m.load_from(cmd)
        return m

    def __str__(self):
        return 'WalkingPadCurStatus(dist=%s, time=%s, steps=%s, speed=%s, state=%s, ' \
               'mode=%s, app_speed=%s, button=%s, rest=%s)' \
               % (self.dist / 100, self.time, self.steps, self.speed / 10, self.belt_state,
                  self.manual_mode, self.app_speed / 30 if self.app_speed > 0 else 0, self.manual_mode,
                  binascii.hexlify(bytearray([self.raw[15], self.raw[17]])).decode('utf8'))


class WalkingPadLastStatus:
    def __init__(self):
        self.raw = None
        self.dist = 0
        self.time = 0
        self.steps = 0
        self.rtime = 0

    def load_from(self, cmd):
        self.raw = bytearray(cmd)
        self.time = WalkingPad.byte2int(cmd[8:])
        self.dist = WalkingPad.byte2int(cmd[11:])
        self.steps = WalkingPad.byte2int(cmd[14:])
        self.rtime = time.time()

    @staticmethod
    def check_type(cmd):
        return bytes(cmd[0:2]) == bytes([248, 167])

    @staticmethod
    def from_data(cmd):
        if not WalkingPadLastStatus.check_type(cmd):
            raise ValueError('Incorrect message type, could not parse')
        m = WalkingPadLastStatus()
        m.load_from(cmd)
        return m

    def __str__(self):
        return 'WalkingPadLastStatus(dist=%s, time=%s, steps=%s, rest=%s)' \
               % (self.dist / 100, self.time, self.steps, binascii.hexlify(self.raw[2:8]).decode('utf8'))


class Controller:
    def __init__(self, address=None, do_read_chars=True):
        self.address = address
        self.do_read_chars = do_read_chars
        self.log_messages_info = True

        self.char_fe01 = None
        self.char_fe02 = None
        self.client = None
        self.last_raw_cmd = None
        self.last_cmd_time = None
        self.last_status = None
        self.last_record = None
        self.minimal_cmd_space = 0.69

        self.handler_cur_status = None
        self.handler_last_status = None
        self.handler_message = None

    async def __aenter__(self):
        await self.run()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()

    def notif_handler(self, sender, data):
        logger_fnc = logger.info if self.log_messages_info else logger.debug
        logger_fnc('Msg: %s' % (', '.join('{:02x}'.format(x) for x in data)))
        already_notified = False

        if WalkingPadCurStatus.check_type(data):
            m = WalkingPadCurStatus.from_data(data)
            self.last_status = m
            already_notified = True
            self.on_cur_status_received(sender, m)
            if self.handler_cur_status:
                self.handler_cur_status(sender, m)
            logger_fnc('Status: %s' % (m,))

        elif WalkingPadLastStatus.check_type(data):
            m = WalkingPadLastStatus.from_data(data)
            self.last_record = None
            already_notified = True
            self.on_last_status_received(sender, m)
            if self.handler_last_status:
                self.handler_last_status(sender, m)
            logger_fnc('Record: %s' % (m,))

        self.on_message_received(sender, data, already_notified)
        if self.handler_message:
            self.handler_message(sender, data, already_notified)

    def on_message_received(self, sender, data, already_notified=False):
        """Override to use as message callback"""

    def on_cur_status_received(self, sender, status: WalkingPadCurStatus):
        """Override to receive current status"""

    def on_last_status_received(self, sender, status: WalkingPadLastStatus):
        """Override to receive last status"""

    def fix_crc(self, cmd):
        return WalkingPad.fix_crc(cmd)

    async def disconnect(self):
        logger.info("Disconnecting")
        await self.client.disconnect()

    async def connect(self, address=None):
        address = address or self.address
        if not address:
            raise ValueError('No address given to connect to')

        logger.info("Connecting to %s" % (address,))
        kwargs = Scanner.get_bleak_kwargs()
        self.client = BleakClient(address, **kwargs)
        return await self.client.connect(timeout=10.0, **kwargs)

    async def send_cmd(self, cmd):
        self.fix_crc(cmd)
        if self.last_cmd_time and time.time() - self.last_cmd_time < self.minimal_cmd_space:
            to_sleep = max(0, min(time.time() - self.last_cmd_time, self.minimal_cmd_space))
            await asyncio.sleep(to_sleep)

        return await self.send_cmd_raw(cmd)

    async def send_cmd_raw(self, cmd):
        self.last_raw_cmd = cmd
        self.last_cmd_time = time.time()
        r = await self.client.write_gatt_char(self.char_fe02, cmd)
        return r

    async def switch_mode(self, mode: int):
        cmd = bytearray([247, 162, 2, mode, 0xff, 253])
        return await self.send_cmd(cmd)

    async def change_speed(self, speed: int):
        cmd = bytearray([247, 162, 1, speed, 0xff, 253])
        return await self.send_cmd(cmd)

    async def stop_belt(self):
        return await self.change_speed(0)

    async def start_belt(self):
        cmd = bytearray([247, 162, 4, 1, 0xff, 253])
        return await self.send_cmd(cmd)

    async def ask_profile(self, profile_idx=0):
        cmd = bytearray(WalkingPad.PAYLOADS_255[profile_idx])
        return await self.send_cmd(cmd)

    async def ask_stats(self):
        cmd = bytearray([247, 162, 0, 0, 162, 253])
        return await self.send_cmd(cmd)

    async def ask_hist(self, mode=0):
        cmd = bytearray([247, 167, 170, 255, 80, 253] if mode == 0 else [247, 167, 170, 0, 81, 253])
        return await self.send_cmd(cmd)

    async def cmd_162_3_7(self, mode=0):
        cmd = bytearray([247, 162, 3, 7, 172, 253])
        return await self.send_cmd(cmd)

    async def set_pref_arr(self, key: int, arr):
        cmd = bytearray([247, 166, key, *arr, 172, 253])
        return await self.send_cmd(cmd)

    async def set_pref_int(self, key: int, val: int, stype: int = 0):
        arr = [stype, *WalkingPad.int2byte(val)]
        return await self.set_pref_arr(key, arr)

    async def set_pref_max_speed(self, speed):
        return await self.set_pref_int(WalkingPad.PREFS_MAX_SPEED, speed)

    async def set_pref_start_speed(self, speed):
        return await self.set_pref_int(WalkingPad.PREFS_START_SPEED, speed)

    async def set_pref_inteli(self, enabled=False):
        return await self.set_pref_int(WalkingPad.PREFS_START_INTEL, int(enabled))

    async def set_pref_sensitivity(self, sensitivity=3):  # 1 = high, 2 = medium, 3 = low
        return await self.set_pref_int(WalkingPad.PREFS_SENSITIVITY, sensitivity)

    async def set_pref_display(self, bit_mask: int):  # 7bits
        return await self.set_pref_int(WalkingPad.PREFS_DISPLAY, bit_mask)

    async def set_pref_child_lock(self, enabled=False):
        return await self.set_pref_int(WalkingPad.PREFS_CHILD_LOCK, int(enabled))

    async def set_pref_units_miles(self, enabled=False):
        return await self.set_pref_int(WalkingPad.PREFS_UNITS, int(enabled))

    async def set_pref_target(self, target_type: int = 0, value: int = 0):
        return await self.set_pref_int(WalkingPad.PREFS_TARGET, value, target_type)

    async def run(self, address=None):
        await self.connect(address)
        client = self.client

        x = client.is_connected
        logger.info("Connected: {0}".format(x))

        self.char_fe01 = None
        self.char_fe02 = None

        for service in client.services:
            logger.info("[Service] {0}: {1}".format(service.uuid, service.description))
            for char in service.characteristics:
                value = None
                if "read" in char.properties:
                    try:
                        if self.do_read_chars and char.uuid != '0000fe01-0000-1000-8000-00805f9b34fb':
                            value = bytes(await client.read_gatt_char(char.uuid))
                    except Exception as e:
                        logger.info('read failed for %s' % (char.uuid,))
                        value = str(e).encode()

                logger.info(
                    "\t[Characteristic] {0}: (Handle: {1}) ({2}) | Name: {3}, Value: {4} ".format(
                        char.uuid,
                        char.handle,
                        ",".join(char.properties),
                        char.description,
                        value,
                    )
                )

                if char.uuid.startswith('0000fe01'):
                    self.char_fe01 = char

                if char.uuid.startswith('0000fe02'):
                    self.char_fe02 = char

                for descriptor in char.descriptors:
                    value = await client.read_gatt_descriptor(descriptor.handle)
                    logger.info(
                        "\t\t[Descriptor] {0}: (Handle: {1}) | Value: {2} ".format(
                            descriptor.uuid, descriptor.handle, bytes(value)
                        )
                    )

        try:
            logger.info('Enabling notification for %s' % (self.char_fe01.uuid,))
            await client.start_notify(self.char_fe01.uuid, self.notif_handler)

        except Exception as e:
            logger.warning("Notify failed: %s" % (e,))

        logger.info('Service enumeration done')






import functools
import itertools
import os
import codecs
from operator import methodcaller, sub


"""
https://stackoverflow.com/questions/2301789/how-to-read-a-file-in-reverse-order
"""


def ceil_div(num, denom):
    return -(-num // denom)


def split(string, separator, keep_separator):
    """
    Splits given string by given separator.
    """
    parts = string.split(separator)
    if keep_separator:
        *parts, last_part = parts
        parts = [part + separator for part in parts]
        if last_part:
            return parts + [last_part]
    return parts


def read_batch_from_end(byte_stream, size, end_position):
    """
    Reads batch from the end of given byte stream.
    """
    if end_position > size:
        offset = end_position - size
    else:
        offset = 0
        size = end_position
    byte_stream.seek(offset)
    return byte_stream.read(size)


def reverse_binary_stream(byte_stream, batch_size=None,
                          lines_separator=None,
                          keep_lines_separator=True):
    if lines_separator is None:
        lines_separator = (b'\r', b'\n', b'\r\n')
        lines_splitter = methodcaller(str.splitlines.__name__,
                                      keep_lines_separator)
    else:
        lines_splitter = functools.partial(split,
                                           separator=lines_separator,
                                           keep_separator=keep_lines_separator)
    stream_size = byte_stream.seek(0, os.SEEK_END)
    if batch_size is None:
        batch_size = stream_size or 1
    batches_count = ceil_div(stream_size, batch_size)
    remaining_bytes_indicator = itertools.islice(
            itertools.accumulate(itertools.chain([stream_size],
                                                 itertools.repeat(batch_size)),
                                 sub),
            batches_count)
    try:
        remaining_bytes_count = next(remaining_bytes_indicator)
    except StopIteration:
        return

    def read_batch(position):
        result = read_batch_from_end(byte_stream,
                                     size=batch_size,
                                     end_position=position)
        while result.startswith(lines_separator):
            try:
                position = next(remaining_bytes_indicator)
            except StopIteration:
                break
            result = (read_batch_from_end(byte_stream,
                                          size=batch_size,
                                          end_position=position)
                      + result)
        return result

    batch = read_batch(remaining_bytes_count)
    segment, *lines = lines_splitter(batch)
    yield from reversed(lines)
    for remaining_bytes_count in remaining_bytes_indicator:
        batch = read_batch(remaining_bytes_count)
        lines = lines_splitter(batch)
        if batch.endswith(lines_separator):
            yield segment
        else:
            lines[-1] += segment
        segment, *lines = lines
        yield from reversed(lines)
    yield segment


def reverse_file(file, batch_size=None,
                 lines_separator=None,
                 keep_lines_separator=True):
    encoding = file.encoding
    if lines_separator is not None:
        lines_separator = lines_separator.encode(encoding)
    yield from map(functools.partial(codecs.decode,
                                     encoding=encoding),
                   reverse_binary_stream(
                           file.buffer,
                           batch_size=batch_size,
                           lines_separator=lines_separator,
                           keep_lines_separator=keep_lines_separator))

